/*
 *  Dimitri KRAVTCHUK (c) 1995-2020
 *  dimitri (dot) kravtchuk (at) free (dot) fr
 *
 *  ...   : 1997... - long dev progress ;-)
 *  v.2.5 : 2015.12 - aligned with STAT-service v.5.0
 *  v.2.6 : 2016.01 - various code remastering..
 *  v.2.7 : 2020.07 - report BRK_ALARM & BRK_PIPE + code remastering..
 *  v.2.8 : 2020.08 - better BRK_ALARM management & diagnostic..
 */

#define VERSION           "(dim) v.2.8"

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <time.h>
#include <sys/time.h>           /* struct timeval */
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define NO_CONN     "STAT *** NO CONNECTION\n"
#define BAD_CMD     "STAT *** BAD COMMAND"        // key prefix for bad command..
#define BRK_ALARM   "STAT *** BREAK-ALARM\n"
#define BRK_PIPE    "STAT *** BREAK-PIPE\n"

struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;

extern int errno;

static int fd;              /* fd of network socket */
short BreakAlarm = 0;       /* Alarm BreakConnection */
short GotAlarm = 0;         // Alarm flag

void err( char *s, int sst );
void msg( char *s );
int Nread();
int Nwrite();


void sigpipe() {}

// >>: alarm..
void sigalrm() {
  signal( SIGALRM, sigalrm );

  if( BreakAlarm ) {
    GotAlarm = 1;
    close( fd );
    Nwrite( 1, BRK_ALARM, sizeof( BRK_ALARM ));
  }
}


void Usage( char *cmd ) {
  fprintf( stderr,
  "%s %s \n"
  "Usage: STATcmd [-k] [-t numb] [-b nsec] -h host [-p port] -c command \n"
  "  -b nsec   -- break connection after nsec-seconds of inactivity\n"
  "  -t numb   -- try to connect only numb times (default: 0 (unlimited)) \n"
  "  -p port   -- STAT-service port number (default: 5000)\n"
  "  -k        -- keep alive\n",
  cmd, VERSION
  );

  exit( -2 );
}


int main( int argc, char *argv[] ) {
  short KeepAlive= 0;
  short port= 5000;               /* TCP port number */
  char *cmd= NULL;                /* STAT command */
  char *host= NULL;               /* ptr to name of host */
  struct hostent *addr;
  unsigned long addr_tmp;
  int i, cnt, sst, nb;
  char buff[1024];
  long T0, T1;
  int Try= 0;

  if( argc < 3 )  Usage( argv[0] );

  for( i= 1; i < argc; i++ ) {
    if( strcmp( argv[i], "-k" ) == 0 )  { KeepAlive= 1;                   continue; }
    if( strcmp( argv[i], "-t" ) == 0 )  { Try= atoi( argv[++i] );         continue; }
    if( strcmp( argv[i], "-p" ) == 0 )  { port= atoi( argv[++i] );        continue; }
    if( strcmp( argv[i], "-b" ) == 0 )  { BreakAlarm= atoi( argv[++i] );  continue; }
    if( strcmp( argv[i], "-h" ) == 0 )  { host= argv[++i];                continue; }
    if( strcmp( argv[i], "-c" ) == 0 )  { cmd= argv[++i];                 continue; }

    Usage( argv[0] );
  }

  if( BreakAlarm < 0 ) BreakAlarm= 0;
  if( host == NULL || cmd == NULL || strlen(cmd) == 0 )  Usage( argv[0] );
  bzero( (char *)&sinhim, sizeof(sinhim) );

  if( atoi(host) > 0 ) {
    /* Numeric */
    sinhim.sin_family = AF_INET;
    sinhim.sin_addr.s_addr = inet_addr( host );
  } else {
    if( (addr= gethostbyname( host )) == NULL )   err( "bad hostname", -1 );
    sinhim.sin_family = addr->h_addrtype;
    bcopy( addr->h_addr, (char*)&addr_tmp, addr->h_length );
    sinhim.sin_addr.s_addr= addr_tmp;
  }

  sinhim.sin_port = htons(port);
  sinme.sin_port = 0;             /* free choice */
  sinme.sin_family = AF_INET;

  time( &T0 );

  for( nb= 0; ; ) {
    signal( SIGPIPE, sigpipe );
    signal( SIGALRM, sigalrm );

    if( BreakAlarm )  alarm( BreakAlarm );
    if( (fd= socket( AF_INET, SOCK_STREAM, 0)) < 0 )    err( "socket", fd );
    if( (sst= bind( fd, (struct sockaddr *)&sinme, sizeof(sinme) )) < 0 )   err( "bind", sst );

    if( connect( fd, (struct sockaddr *)&sinhim, sizeof(sinhim) ) < 0 ) {
      close( fd );
      Nwrite( 1, NO_CONN, sizeof( NO_CONN ) );
      if( Try && ++nb >= Try )  exit(0);

      sleep( 20 );
      continue;
    }

    nb= 0;
    time( &T1 );

    // >>: connection..
    sprintf( buff, "STAT *** OK CONNECTION %ld sec.\n", T1-T0 );
    Nwrite( 1, buff, strlen( buff ) );

    // >>: send command..
    Nwrite( fd, cmd, strlen(cmd) );

    // >>: get OK..
    if( (cnt= Nread( fd, buff, sizeof(buff) )) <= 0 ) {
      if( GotAlarm ) {
        if( !KeepAlive )  exit(0);

        GotAlarm= 0;
        sleep( 20 );
        continue;
      }

      close( fd );
      Nwrite( 1, BRK_PIPE, sizeof( BRK_PIPE ) );
      continue;
    }

    Nwrite( 1, buff, strlen( buff ) );
    if( strncmp( buff, BAD_CMD, sizeof( BAD_CMD ) ) == 0 )   break;  // bad command..

    // >>: loop on command output..
    while( 1 ) {
      time( &T0 );   /* timestamp to get inactivity info */

      if( (cnt= Nread( fd, buff, sizeof(buff) )) <= 0 ) {
        if( !KeepAlive ) {
          close( fd );
          exit( 0 );
        }

        if( !GotAlarm ) {
          close( fd );
          Nwrite( 1, BRK_PIPE, sizeof( BRK_PIPE ) );
        }

        GotAlarm= 0;
        sleep( 20 );
        break;
      }

      Nwrite( 1, buff, strlen( buff ) );
    }
  }

  close( fd );
  exit( 0 );
}


void err( char *msg, int sst ) {
  fprintf( stderr, "## STATcmd pid-%d> %s sst: %d errno: %d\n", getpid(), msg, sst, errno );
  perror( msg );
  exit( -2 );
}


void msg( char *msg) {
  fprintf( stderr, "## STATcmd pid-%d> %s\n", getpid(), msg );
}


int Nread( int fd, char *buf, int count ) {
  int cnt;

  if( BreakAlarm )  alarm( BreakAlarm );  // reset Alarm before read()..
  cnt= read( fd, buf, count );
  if( cnt >= 0 ) buf[cnt]= '\0';

  return( cnt );
}


int Nwrite( int fd, char *buf, int count ) {
  int sst;

  if( (sst= write( fd, buf, count )) <= 0 ) {
    close( fd );

    if( fd == 1 )  err( "Nwrite error to Client..", sst );
    else           err( "Nwrite error to STATsrv..", sst );
  }

  return( sst );
}
