/*
 * STATsrv: STAT-service daemon
 *          part of dim_STAT tool <http://dimitrik.free.fr>
 *
 * (c) 1997-2021 by (dim) Dimitri KRAVTCHUK <dimitri.kravtchuk <at> free.fr>
 *
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   *
 ***************************************************************************
 */


/*
    ...   : -- 1997 ... : long dev progress.. ;-)
    v.3.9 : -- Oct.2008 : fixed kill chld processes!
    v.3.10: -- May.2009 : added "&" to hidden command check!
                        : fixed possible data oferflow in cmd sizeof()!!
    v.3.11: -- Jun.2009 : keep only alphanumeric symbols within a command line!!
    v.3.12: --          : updated accepted chars: see check_char() function..
    v.4.0 : -- Mar.2010 : extended syntax statname:port
    v.4.1 : -- Nov.2010 : reason of EXIT CMD in the message
    v.5.0 : -- Nov.2015 : allowing to use a different home directory and access file
                        :    -d   - new option for the home directory
                        :    -a   - access file name
                        : auto-set env variables: STAT_HOME & STAT_PORT
                        : allowing a use of $(STAT_HOME) & $(STAT_PORT) in access file
    v.5.1 : -- Jan.2016 : various code remastering..
    v.5.2 : -- Jul.2020 : more code remastering..
    v.5.3 : -- Sep.2020 : kill chld "softly" & report #killed chld..
    v.5.4 : -- Sep.2020 : add [-b] BreakAlarm signal + cleanup()..
    v.5.5 : -- May.2021 : allowing also $VAL replace..
    v.5.6 : -- Aug.2021 : MAX_PORT is 65000 now..
    v.5.7 : -- Sep.2021 : add support for "include file" in access..
    v.5.8 : -- Jan.2023 : SOCKET reuse options..
    v.5.9 : -- Jan.2023 : proper "forked" on-signal cleanup..
*/

#define VERSION           "(dim) v.5.9"

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <time.h>
#include <sys/time.h>           /* struct timeval */

struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;

extern int errno;
extern void exit();
void err( char *s, int sst );
void print_msg();
int Nread();
int Nwrite();
int kill_chld();

#define CMDLEN 128
#define P_SIZE 256        // PID list size (kill child)
#define B_SIZE 512        // buffer on reading PID (kill child)

#define MAX_PORT          65000
#define DEF_PORT          5000
#define DEF_ALARM         120
#define DEF_HOME          "/etc/STATsrv"
#define DEF_ACCESS        "access"

#define KEY_FULLCMD_SIZE  512
#define ACCESS_NAME_SIZE  512
#define ACCESS_BUFF_SIZE  2048  //max allowed line length in access file

#define ENV_BUFF_SIZE     1024
#define ENV_STAT_HOME     "SRV_HOME"
#define ENV_STAT_PORT     "SRV_PORT"

// global vars
char ACCESS_NAME[ ACCESS_NAME_SIZE ];
char *Home_DIR= DEF_HOME;

// need to be global to allow a clean exit on signal..
int fd= -1;                         /* fd of network socket */
int fd_cmd= -1;
int forked= 0;
char CMD[ CMDLEN ];

// >>: cleanup..
void cleanup( char* msg ) {
  time_t T;
  int chld;

  //kill all chld procs!!!
  chld= kill_chld();

  if( fd_cmd != -1 )          {  shutdown( fd_cmd, SHUT_RDWR ); close( fd_cmd ); }
  if( !forked && fd != -1 )   {  shutdown( fd, SHUT_RDWR ); close( fd ); }

  time( &T );

  if( strlen( CMD) > 0 ) {
    fprintf( stderr, "STATsrv-%d> Exit    -> CMD= [%s] (%s) killed:%d TIME: %s",
      getpid(), CMD, msg, chld, ctime(&T) );
    fflush( stderr );
  } else {
    fprintf( stderr, "STATsrv-%d> Exit    -> (%s) killed:%d TIME: %s",
      getpid(), msg, chld, ctime(&T) );
    fflush( stderr );

    kill( 0, SIGTERM );
  }

  exit(0);
}


// >>: signal traps..
void sigalarm() { cleanup( "EXIT by Break-ALARM" ); }
void sigterm()  { cleanup( "EXIT by SIGTERM" ); }
void sigpipe()  { signal(SIGPIPE, sigpipe); }
void sigchld()  { int sst; wait( &sst ); signal( SIGCHLD, sigchld ); }


// >>: Usage
void Usage( char *cmd ) {
  fprintf( stderr,
    "%s %s\nUsage: %s [-p [IP:]port (def:%d)] [-d home_dir (def:%s)] [-a access_file (def:%s)] [-b BreakAlarm (def:%d)] \n",
    cmd, VERSION, cmd, DEF_PORT, DEF_HOME, DEF_ACCESS, DEF_ALARM );

  exit( -2 );
}


void err( char *s, int sst ) {
  fprintf( stderr, "STATsrv-%d> %s sst: %d errno: %d\n", getpid(), s, sst, errno );
  perror( s );
  fflush( stderr );
  exit( -2 );
}


void print_msg( char *s ) {
  fprintf( stderr, "STATsrv-%d> %s", getpid(), s );
  if( errno != 0 )  fprintf( stderr, " errno: %d", errno );
  fprintf( stderr, "\n" );
  fflush( stderr );
}


int Nread( int fd, char *buf, int count ) {
  int cnt;
  cnt= read( fd, buf, count );
  if( cnt >= 0 ) buf[cnt]= '\0';
  return( cnt );
}


int Nwrite( int fd, char *buf, int count ) {
  return( write( fd, buf, count ) );
}


// Valid chars within a shell command
int check_char( char ch ) {
  static char valid[]= ".,:*/-_ ";
  int i;

  if( isalnum( ch ) )  return( 0 );

  for( i= 0; valid[i] != '\0'; i++ )
    if( valid[i] == ch )  return( 0 );

  return( -1 );
}


// replace
int str_replace( char *buff, char* old, char* new ) {
  int i, oldlen, newlen;
  char out[ ACCESS_BUFF_SIZE ], *chp;

  oldlen= strlen( old );
  newlen= strlen( new );
  out[0]= '\0';

  for( i= 0, chp= buff; *chp != '\0'; ) {
    if( i + newlen >= ACCESS_BUFF_SIZE )  break;

    if( memcmp( chp, old, oldlen ) == 0 ) {
      memcpy( &out[i], new, newlen );
      i += newlen;
      chp += oldlen;
      continue;
    }

    out[i++]= *chp++;
  }

  out[i]= '\0';
  strcpy( buff, out );

  return( 0 );
}


// >>: Find the matching shell command..
int find_command( char *fname, char* cmd, char *key_fullcmd, char *addr, int port, int fd ) {
  FILE *fp;
  char msg[200], *chp;
  char key[32], key_cmd[ CMDLEN ], buff[ ACCESS_BUFF_SIZE ];
  int access, Ok, found;

  if( (fp= fopen( fname, "r" )) == NULL )  return( 0 );

  for( access= found= 0, Ok= 1; fgets( buff, sizeof(buff), fp ) != NULL; ) {
    if( buff[0] == '#' )  continue;
    if( strlen( buff ) < 5 )  continue;

    str_replace( buff, "\t", "" );
    str_replace( buff, "$(SRV_HOME)", Home_DIR );
    str_replace( buff, "$SRV_HOME", Home_DIR );
    sprintf( msg, "%d", port );
    str_replace( buff, "$(SRV_PORT)", msg );
    str_replace( buff, "$SRV_PORT", msg );

    // let's limit sscanf ;-)
    // note: if the params were intentionally given too long, then we'll read some trash ;-)
    //       but it'll not overflow any buffer..
    key[0]= key_cmd[0]= key_fullcmd[0]= '\0';
    sprintf( msg, "%%%ds%%%ds%%%ds", (int)sizeof(key)-1, (int)sizeof(key_cmd)-1, KEY_FULLCMD_SIZE-1 );
    sscanf( buff, msg, key, key_cmd, key_fullcmd );

    if( strcmp( key, "access" ) == 0 ) {
      if( access && Ok )  continue;
      access= 1;
      Ok= 0;
      if( strcmp( addr, key_cmd ) == 0 )  Ok= 1;
      continue;
    }

    if( strcmp( key, "include" ) == 0 ) {
      access= 0;
      if( !Ok )  continue;

      strcpy( buff, key_cmd );
      found= find_command( buff, cmd, key_fullcmd, addr, port, fd );

      if( found )  break;
      continue;
    }

    if( strcmp( key, "command" ) == 0 ) {
      access= 0;
      if( !Ok )  continue;

      // check the case cmd:port
      if( (chp= (char *)strchr( key_cmd, ':' )) != NULL ) {
        if( atoi( chp+1 ) != port )  continue;
        *chp= '\0';
      }

      if( fd ) {
        sprintf( msg, "STAT: %s\n", key_cmd );
        Nwrite( fd, msg, strlen( msg ) );
        continue;
      }

      if( strcmp( cmd, key_cmd ) == 0 )  { found= 1; break; }
    }
  }

  fclose( fp );
  return( found );
}


// >>: Check and Execute a shell command
FILE *check_command( char* cmdline, char* addr, int port, int fd ) {
  FILE *fp;
  char cmd[ CMDLEN ], msg[200], buff[ ACCESS_BUFF_SIZE ], *chp, *tail;
  char key_fullcmd[ KEY_FULLCMD_SIZE ];
  int i, n, found, List= 0;

  for( cmd[0]= '\0', i= n= 0; i < (sizeof( cmd ) - 2) && cmdline[n] != '\0'; n++ ) {
    if( cmdline[n] == ' ' ) {
      if( i > 0 )  break;
      continue;
    }

    cmd[i++]= cmdline[n];
  }

  cmd[i]= '\0';
  tail= &cmdline[n];

  // remove a path from name
  while( (chp= (char *)strchr( cmd, '/' )) != NULL )  strcpy( cmd, chp+1 );

  if( strcmp( cmd, "STAT_LIST" ) == 0 ) {
    sprintf( msg, "STAT *** LIST COMMAND (%s)\n", cmd );
    Nwrite( fd, msg, strlen( msg ) );
    List= 1;
  }

  found= find_command( ACCESS_NAME, cmd, &key_fullcmd[0], addr, port, (List ? fd : 0) );

  if( List ) {
    sprintf( msg, "STAT *** LIST END (%s)\n", cmd );
    Nwrite( fd, msg, strlen( msg ) );
    return( NULL );
  }

  if( !found ) {
    sprintf( msg, "STAT *** BAD COMMAND (no access: %s)\n", cmd );
    Nwrite( fd, msg, strlen( msg ) );
    return( NULL );
  }

  /* avoid buff overflow */
  strcpy( buff, key_fullcmd );
  strcat( buff, " " );
  strncat( buff, tail, sizeof(buff) - strlen(buff) - 1 );
  buff[ sizeof(buff)-1 ]= '\0';

  if( (fp= popen( buff, "r" )) == NULL ) {
    sprintf( msg, "STAT *** BAD COMMAND (cannot start: %s)\n", cmd );
    Nwrite( fd, msg, strlen( msg ) );
    return( fp );
  }

  if( fgets( buff, sizeof(buff), fp ) == NULL ) {
    sprintf( msg, "STAT *** BAD COMMAND (not found: %s)\n", cmd );
    Nwrite( fd, msg, strlen( msg ) );
    pclose( fp );
    return( NULL );
  }

  sprintf( msg, "STAT *** OK COMMAND (cmd: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  Nwrite( fd, buff, strlen( buff ) );
  return( fp );
}


/*
 *  In many cases child processes may still continue their execution
 *  even(!) if command pipe is broken...
 *  Need an explicit kill here to be sure to stop all stuff..
 */

// >>: kill_chld
int kill_chld() {
  FILE *pp;
  char buff[ B_SIZE + 8];
  long pid;
  long plist[ P_SIZE ];
  int  n, i;

  pid= getpid();
  sprintf( buff, "%s/bin/chldPID.sh", Home_DIR );

  if( access( buff, 0 ) == 0 ) {
    sprintf( buff, "%s/bin/chldPID.sh %ld", Home_DIR, pid );
    pp= popen( buff, "r" );

    for( n= 0; n < P_SIZE && fgets( buff, B_SIZE, pp ) != NULL; n++ ) {
      sscanf( buff, "%ld", &plist[n] );
      if( plist[n] < 2 )  n--;  // skip "master" UNIX pid(s)
    }

    pclose( pp );

    // kill-me "softly"..
    for( i= 0; i < n && i < P_SIZE; i++ )  kill( plist[i], SIGTERM );
    sleep( 3 );
    for( i= 0; i < n && i < P_SIZE; i++ )  kill( plist[i], SIGKILL );

    return( n );
  }

  return( 0 );
}


// >>: STATsrv..
void STATsrv( int fd, char *addr, int port, int tm ) {
  time_t T;
  int cnt, chld, Nerr;
  char *chp;
  char buff[1024];
  FILE *pp;

  Nerr= errno= chld= 0;
  if( tm )  alarm( tm );
  if( (cnt= Nread( fd, buff, sizeof(buff) )) <= 0 )  err( "Nread", cnt );

  // replace any "suspected" characters whithin a buff command line by spaces
  // to avoid any other hidden order...
  for( chp= buff; *chp != '\0'; chp++ )
    if( check_char( *chp ) != 0 )
      *chp= ' ';

  strncpy( CMD, buff, CMDLEN );
  CMD[ CMDLEN-1 ]= '\0';

  time( &T );
  fprintf( stderr, "STATsrv-%d:%d> Connect -> [%s] CMD= [%s] TIME: %s",
    getpid(), port, addr, CMD, ctime(&T) );
  fflush( stderr );

  if( (pp= check_command( buff, addr, port, fd )) != NULL ) {
    if( tm )  alarm( tm );

    while( fgets( buff, sizeof( buff ), pp ) != NULL ) {
      if( Nwrite( fd, buff, strlen( buff ) ) <= 0 ) {
        // lost connection..
        Nerr= 1;
        break;
      }

      if( tm )  alarm( tm );
    }
    // kill child procs!!
    chld= kill_chld();
    pclose( pp );
  }

  close( fd );
  time( &T );
  fprintf( stderr, "STATsrv-%d:%d> Exit    -> [%s] CMD= [%s] (%s) killed:%d TIME: %s",
    getpid(), port, addr, CMD, Nerr ? "lost connection" : "broken pipe", chld, ctime(&T) );
  fflush( stderr );
}


//----------------------------------------------------------------------------------
// >>: Main start...
//----------------------------------------------------------------------------------

int main( int argc, char *argv[] ) {
  static char env_buff[ ENV_BUFF_SIZE ];  /* putenv() needs a static buffer      */
  int port= DEF_PORT;                   /* TCP port number */
  int domain, fromlen;
  int i, len, sst;
  char *chp;
  char *IP= NULL;
  char *access_file= DEF_ACCESS;
  int tm= DEF_ALARM;

  struct sockaddr_in peer;
  int peerlen= sizeof(peer);
  CMD[0]= '\0';

  for( i= 1; i < argc; i++ ) {
    if( strcmp( argv[i], "-b" ) == 0 ) {
      tm= atoi( argv[++i] );
      continue;
    }

    if( strcmp( argv[i], "-d" ) == 0 ) {
      Home_DIR= argv[++i];
      continue;
    }

    if( strcmp( argv[i], "-a" ) == 0 ) {
      access_file= argv[++i];
      continue;
    }

    if( strcmp( argv[i], "-p" ) == 0 ) {
      if( (chp= (char *)strchr( argv[++i], ':' )) == NULL )
        port= atoi( argv[i] );
      else {
        *chp = '\0';
        IP= argv[i];
        port= atoi( ++chp );
      }

      if( port > MAX_PORT || port < 1 ) {
        fprintf( stderr, "## Bad port number: [%d] (use range 1:%d)\n\n", port, MAX_PORT );
        Usage( argv[0] );
      }

      continue;
    }

    Usage( argv[0] );
  }

  // init start..
  fprintf( stderr, "\nSTAT-service %s ==> PID:%d Init.. \n", VERSION, getpid() );

  if( strlen( Home_DIR ) + strlen( access_file ) + 4 > ACCESS_NAME_SIZE  ) {
    fprintf( stderr, "## Access file path is too long: [%s/%s]\n", Home_DIR, access_file );
    exit( -2 );
  }

  sprintf( ACCESS_NAME, "%s/%s", Home_DIR, access_file );

  if( access( ACCESS_NAME, 0 ) != 0 ) {
    fprintf( stderr, "## Access file not found: [%s]\n", ACCESS_NAME );
    exit( -2 );
  }

  len= 0;
  sprintf( &env_buff[ len ], "%s=%s", ENV_STAT_HOME, Home_DIR );
  putenv( &env_buff[ len ] );

  len += strlen( &env_buff[ len] ) + 2 ;
  sprintf( &env_buff[ len ], "%s=%d", ENV_STAT_PORT, port );
  putenv( &env_buff[ len ] );

  setpgrp();

  sinme.sin_port=  htons(port);
  sinme.sin_family= AF_INET;

  if( IP == NULL ) sinme.sin_addr.s_addr= INADDR_ANY;
  else {
    if( (sinme.sin_addr.s_addr= inet_addr( IP )) == (in_addr_t)-1 ) {
      printf( "## Bad IP address: %s\n\n", IP );
      Usage( argv[0] );
    }
  }

  // ready to start..
  fprintf( stderr, "\nSTAT-service %s ==> PID:%d Start.. \n > IP:[%s] \n > Port:[%d] \n > Home:[%s] \n > Access:[%s] \n > BreakAlarm:[%d] \n",
    VERSION, getpid(), (IP == NULL) ? "any" : IP, port, Home_DIR, access_file, tm );

  if( (fd= socket( AF_INET, SOCK_STREAM, 0 )) < 0 )  err( "socket", fd );

  int opt = 1;

  if( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt) ) < 0 )  err( "socket", fd );
  if( setsockopt( fd, SOL_SOCKET, SO_REUSEPORT, &opt, sizeof(opt) ) < 0 )  err( "socket", fd );

  while( 1 ) {
    if( (sst= bind( fd, (struct sockaddr *)&sinme, sizeof(sinme) )) >= 0 )  break;
    print_msg( "-- Bind Error, will try again in 30sec..." );
    sleep( 30 );
  }

  signal( SIGPIPE, sigpipe );
  signal( SIGCHLD, sigchld );
  signal( SIGTERM, sigterm );
  signal( SIGALRM, sigalarm );

  listen( fd, 0 );   /* allow a queue of 0 */
  print_msg( "*** READY for Service! ***" );

  fromlen= sizeof(frominet);
  domain= AF_INET;

  // >>: main loop..
  while( 1 ) {
    if( (fd_cmd= accept( fd, (struct sockaddr *)&frominet, (socklen_t *)&fromlen )) < 0 ) {
      if( errno == EINTR )  continue;
      err( "accept", fd_cmd );
    }

    if( fork() == 0 ) {
      // >>: child process..
      forked= 1;
      if( (sst= getpeername( fd_cmd, (struct sockaddr *) &peer, (socklen_t *)&peerlen )) < 0 )
        err( "getpeername", sst );

      STATsrv( fd_cmd, inet_ntoa(peer.sin_addr), port, tm );

      shutdown( fd_cmd, SHUT_RDWR );
      close( fd_cmd );

      exit(0);
    }

    close( fd_cmd );
  }
}
